import amncore.compat
import amncore.mixins
import amncore.xcepts
import bombfuse
import copy
import ctypes
import errno
try:
    import http_packet
except ImportError:
    from . import http_packet
import os
import paynspray.rando
import random
import re
import select
import socket
try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer
import struct
import threading

MATCH_IPV6 = r'^\s*((([0-9A-Fa-f]{1,4}:){7}([0-9A-Fa-f]{1,4}|:))|(([0-9A-Fa-f]{1,4}:){6}(:[0-9A-Fa-f]{1,4}|((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(([0-9A-Fa-f]{1,4}:){5}(((:[0-9A-Fa-f]{1,4}){1,2})|:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(([0-9A-Fa-f]{1,4}:){4}(((:[0-9A-Fa-f]{1,4}){1,3})|((:[0-9A-Fa-f]{1,4})?:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9A-Fa-f]{1,4}:){3}(((:[0-9A-Fa-f]{1,4}){1,4})|((:[0-9A-Fa-f]{1,4}){0,2}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9A-Fa-f]{1,4}:){2}(((:[0-9A-Fa-f]{1,4}){1,5})|((:[0-9A-Fa-f]{1,4}){0,3}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9A-Fa-f]{1,4}:){1}(((:[0-9A-Fa-f]{1,4}){1,6})|((:[0-9A-Fa-f]{1,4}){0,4}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(:(((:[0-9A-Fa-f]{1,4}){1,7})|((:[0-9A-Fa-f]{1,4}){0,5}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:)))(%.+)?\s*$'
MATCH_IPV4 = r'^\s*(?:(?:25[0-5]|2[0-4][0-9]|[0-1]?[0-9]{1,2})[.](?:25[0-5]|2[0-4][0-9]|[0-1]?[0-9]{1,2})[.](?:25[0-5]|2[0-4][0-9]|[0-1]?[0-9]{1,2})[.](?:25[0-5]|2[0-4][0-9]|[0-1]?[0-9]{1,2}))\s*$'
MATCH_IPV4_PRIVATE = r'^\s*(?:10\.|192\.168|172.(?:1[6-9]|2[0-9]|3[01])\.|169\.254)'
MATCH_MAC_ADDR = r'^(?:[0-9a-fA-F]{2}([-:]))(?:[0-9a-fA-F]{2}\1){4}[0-9a-fA-F]{2}$'

ipv6_support = None

def supports_ipv6():
    global ipv6_support
    if ipv6_support is not None:
        return ipv6_support

    ipv6_support = False

    if hasattr(socket, "AF_INET6") == True:
        try:
            s = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            s.close()
            ipv6_support = True
        except Exception as e:
            pass

    return (ipv6_support and socket.has_ipv6)

class SocketParams(object):
    def __init__(self, **kwargs):
        if 'PeerHost' in kwargs:
            self.peerhost = kwargs['PeerHost']
        elif 'PeerAddr' in kwargs:
            self.peerhost = kwargs['PeerAddr']
        else:
            self.peerhost = None
            
        if 'LocalHost' in kwargs:
            self.localhost = kwargs['LocalHost']
        elif 'LocalAddr' in kwargs:
            self.localhost = kwargs['LocalAddr']
        else:
            self.localhost = '0.0.0.0'
            
        if 'PeerPort' in kwargs and kwargs['PeerPort'] is not None:
            self.peerport = int(kwargs['PeerPort'])
        else:
            self.peerport = 0
        
        if 'LocalPort' in kwargs and kwargs['LocalPort'] is not None:
            self.localport = int(kwargs['LocalPort'])
        else:
            self.localport = 0
            
        if 'Bare' in kwargs:
            self.bare = kwargs['Bare']
        else:
            self.bare = False
            
        if 'SSL' in kwargs:
            self.ssl = kwargs['SSL']
        else:
            self.ssl = False
            
        if 'SSLContext' in kwargs:
            self.sslctx = kwargs['SSLContext']
        else:
            self.sslctx = None
            
        SSLv2 = 1
        SSLv3 = 2
        SSLv23 = 3
        TLSv1 = 4
        TLSv1_1 = 5
        TLSv1_2 = 6

        supported_ssl_versions = ['Auto', 'SSL2', 'SSL23', 'TLS1', 'SSL3', SSLv2, SSLv3, SSLv23, TLSv1]
        if 'SSLVersion' in kwargs and kwargs['SSLVersion'] in supported_ssl_versions:
            self.ssl_version = kwargs['SSLVersion']
        else:
            self.ssl_version = None
            
        supported_ssl_verifiers = ['CLIENT_ONCE', 'FAIL_IF_NO_PEER_CERT', 'NONE', 'PEER']
        if 'SSLVerifyMode' in kwargs and kwargs['SSLVerifyMode'] in supported_ssl_verifiers:
            self.ssl_verify_mode = kwargs['SSLVerifyMode']
        else:
            self.ssl_verify_mode = None

        if 'SSLCompression' in kwargs:
            self.ssl_compression = kwargs['SSLCompression']
        else:
            self.ssl_compression = None

        if 'SSLCipher' in kwargs:
            self.ssl_cipher = kwargs['SSLCipher']
        else:
            self.ssl_cipher = None
            
        if 'SSLCert' in kwargs and kwargs['SSLCert'] is not None and os.path.isfile(kwargs['SSLCert']) == True:
            try:
                fd = file(kwargs['SSLCert'], 'rb')
                self.ssl_cert = fd.read()
                fd.close()
            except Exception as e:
                pass
        else:
            self.ssl_cert = None
            
        if 'SSLClientCert' in kwargs and kwargs['SSLClientCert'] is not None and os.path.isfile(kwargs['SSLClientCert']) == True:
            try:
                fd = file(kwargs['SSLClientCert'], 'rb')
                self.ssl_client_cert = fd.read()
                fd.close()
            except Exception as e:
                pass
        else:
            self.ssl_client_cert = None
            
        if 'SSLClientKey' in kwargs and kwargs['SSLClientKey'] is not None and os.path.isfile(kwargs['SSLClientKey']) == True:
            try:
                fd = file(kwargs['SSLClientKey'], 'rb')
                self.ssl_client_key = fd.read()
                fd.close()
            except Exception as e:
                pass
        else:
            self.ssl_client_key = None

        if 'Proxies' in kwargs and kwargs['Proxies'] is not None:
            self.proxies = map(lambda a: map(lambda b: b.strip(), a.split(':')), map(lambda a: a.strip(), kwargs['Proxies'].split(',')))
        else:
            self.proxies = None
            
        if 'Proto' in kwargs:
            self.proto = kwargs['Proto'].lower()
        else:
            self.proto = 'tcp'
        
        if 'Server' in kwargs:
            self.server = kwargs['Server']
        else:
            self.server = False
        
        if 'Comm' in kwargs and kwargs['Comm'] is not None:
            if issubclass(kwargs['Comm'], SocketComm) == True:
                self.comm = kwargs['Comm']()
            else:
                self.comm = None
        else:
            self.comm = None
            
        if 'Context' in kwargs:
            self.context = kwargs['Context']
        else:
            self.context = {}

        if self.comm is None:
            self.comm = LocalComm()

        if self.server == True and self.proto == 'udp':
            self.server = False

        if 'Retries' in kwargs:
            self.retries = int(kwargs['Retries'])
        else:
            self.retries = 0

        # proposed change for 0.1.5
        if 'Timeout' in kwargs:
            if kwargs['Timeout'] is None:
                # handle None cast
                self.timeout = None
            else:
                # we set a timeout of 0 to None to disable timeout (instead of making socket non-blocking)
                # this value is ONLY used for socket.connect()
                self.timeout = int(kwargs['Timeout'])
                if self.timeout == 0:
                    self.timeout = None
        else:
            self.timeout = 5

        if 'IPv6' in kwargs:
            self.v6 = kwargs['IPv6']
        else:
            self.v6 = False

    @classmethod
    def from_kwargs(self, **kwargs):
        return self(**kwargs)
        
    @property
    def is_server(self):
        return (self.server == True)
        
    @property
    def is_client(self):
        return (self.server == False)
        
    @property
    def is_tcp(self):
        return (self.proto == 'tcp')
        
    @property
    def is_udp(self):
        return (self.proto == 'udp')
        
    @property
    def is_ip(self):
        return (self.proto == 'ip')

    @property
    def is_bare(self):
        return (self.bare == True)

    @property
    def is_ssl(self):
        return (self.ssl == True)
        
    @property
    def is_v6(self):
        return (self.v6)
        
    @property
    def peeraddr(self):
        return self.peerhost
        
    @property
    def localaddr(self):
        return self.localhost
        
def create_sock(**kwargs):
    return create_param(SocketParams.from_kwargs(**kwargs))
    
def create_param(param):
    return param.comm.create(param)
    
def create_tcp(**kwargs):
    kwargs.update({'Proto': 'tcp'})
    return create_param(SocketParams.from_kwargs(**kwargs))
    
def create_tcp_server(**kwargs):
    kwargs.update({'Server': True})
    return create_tcp(**kwargs)
    
def create_udp(**kwargs):
    kwargs.update({'Proto': 'udp'})
    return create_param(SocketParams.from_kwargs(**kwargs))
    
def create_ip(**kwargs):
    kwargs.update({'Proto': 'ip'})
    return create_param(SocketParams.from_kwargs(**kwargs))
    
def is_ipv4(addr):
    if addr is None:
        return False
    return (re.search(MATCH_IPV4, addr) is not None)

def is_ipv6(addr):
    if addr is None:
        return False
    return (re.search(MATCH_IPV6, addr) is not None)
        
def is_mac_addr(addr):
    if addr is None:
        return False
    return (re.search(MATCH_MAC_ADDR, addr) is not None)

def is_dotted_ip(addr):
    if addr is None:
        return False
    return ((supports_ipv6() == True and (re.search(MATCH_IPV6, addr) is not None)) or (re.search(MATCH_IPV4, addr) is not None))

def is_internal(addr):
    if addr is None:
        return False
    return (re.search(MATCH_IPV4_PRIVATE, addr) is not None)
    
def getaddresses(hostname, accept_ipv6 = True):
    if hostname is None:
        return []
    if (is_ipv4(hostname) == True) or ((accept_ipv6 == True) and (is_ipv6(hostname) == True)):
        return [hostname]
        
    addrlist = []
    res = socket.getaddrinfo(hostname, None)
    
    if res is None:
        return []
    
    for i in res:
        if (is_ipv4(i[4][0]) == True) or ((accept_ipv6 == True) and (is_ipv6(i[4][0]) == True)):
            addrlist.append(i[4][0])

    return addrlist
    
def getaddress(hostname, accept_ipv6 = True):
    retval = getaddresses(hostname, accept_ipv6)
    
    if len(retval) > 0:
        return retval[0]
    else:
        return None
    
def getaddrinfo(host, port = 0):
    """
    Wrapper for socket.getaddrinfo() that skips dotted IPv4 resolution
    """
    if is_ipv4(host) == True:
        return [(socket.AF_INET, 0, 0, '', (host, port))]
        
    return socket.getaddrinfo(host, port)
    
def getsockaddr(host, port):
    """
    Returns a tuple as (host, port) for IPv4 hosts and (host, port, flow-info, scope-id) for IPv6 hosts
    """
    if host == '::ffff:0.0.0.0':
        if supports_ipv6() == True:
            host = '::'
        else:
            host = '0.0.0.0'
            
    retval = getaddrinfo(host, port)
    if retval is not None and len(retval) > 0:
        return retval[0][4]
    else:
        return None

def parsesockaddr(saddr):
    """
    Returns a list of [family, host, port] from a Pythonic sock address
    """
    if is_ipv6(saddr[0]) == True:
        af = socket.AF_INET6
        host, port, _, _ = saddr
    elif is_ipv4(saddr[0]) == True:
        af = socket.AF_INET
        host, port = saddr

    return [af, host, port]

def getsrcaddr(dest = '8.8.8.8'):
    """
    Attempts to get the localhost address by establishing a bogus connection
    """
    try:
        s = create_udp(PeerHost = dest, PeerPort = 31337)
        r = s.getsockname()
        s.close()
        
        return r[0]
        
    except Exception as e:
        return '127.0.0.1'

def ipv6_link_address(iface):
    r = getsrcaddr("FF02::1%{}".format(iface))
    if r is None or re.search(r'^fe80', r, re.IGNORECASE) is None:
        return None
    return r
    
class sockaddr(ctypes.Structure):
    _fields_ = [("sa_family", ctypes.c_ushort),
                ("__pad1", ctypes.c_ushort),
                ("ipv4_addr", ctypes.c_byte * 4),
                ("ipv6_addr", ctypes.c_byte * 16),
                ("__pad2", ctypes.c_ulong)]

# little workaround for Windows (fixed in python3)
if hasattr(socket, "inet_pton") == False and amncore.compat.is_windows() == True:
    WSAStringToAddressA = ctypes.windll.ws2_32.WSAStringToAddressA
    WSAAddressToStringA = ctypes.windll.ws2_32.WSAAddressToStringA
    
    # packs an IP string (IPv4 or IPv6) into a binary string
    def inet_pton(address_family, ip_string):
        addr = sockaddr()
        addr.sa_family = address_family
        addr_size = ctypes.c_int(ctypes.sizeof(addr))
        
        if WSAStringToAddressA(ip_string, address_family, None, ctypes.byref(addr), ctypes.byref(addr_size)) != 0:
            raise socket.error(ctypes.FormatError())
            
        if address_family == socket.AF_INET:
            return ctypes.string_at(addr.ipv4_addr, 4)
        elif address_family == socket.AF_INET6:
            return ctypes.string_at(addr.ipv6_addr, 16)
        else:
            raise ValueError("Invalid address family provided")
            
    setattr(socket, "inet_pton", inet_pton)
    
def addr_itoa(addr, v6 = False):
    nboa = addr_iton(addr, v6)
    return addr_ntoa(nboa)
    
def addr_atoi(addr):
    return resolv_nbo_i(addr)
    
def addr_aton(addr):
    return resolv_nbo(addr)
    
def resolv_nbo(host):
    # getaddrinfo returns IPv4 or IPv6 address of host
    res = getaddrinfo(getaddress(host, True))[0]
    if res is not None:
        if res[0] == socket.AF_INET:
            try:
                return str(socket.inet_pton(socket.AF_INET, res[4][0]).decode('latin-1'))
            except UnicodeEncodeError:
                return socket.inet_pton(socket.AF_INET, res[4][0])
        elif res[0] == socket.AF_INET6:
            try:
                return str(socket.inet_pton(socket.AF_INET6, res[4][0]).decode('latin-1'))
            except UnicodeEncodeError:
                return socket.inet_pton(socket.AF_INET6, res[4][0])
        else:
            return ""
    else:
        return ""
    
def resolv_nbo_i(host):
    return addr_ntoi(resolv_nbo(host))
    
def resolv_nbo_list(host):
    return map(lambda addr: resolv_nbo(addr), getaddresses(host))
    
def resolv_nbo_i_list(host):
    return map(lambda addr: addr_ntoi(addr), resolv_nbo_list(host))
    
def resolv_to_dotted(host):
    res = addr_ntoa(addr_aton(host))
    return res
    
def addr_atoc(mask):
    if is_ipv6(mask) == True:
        bits = 128
    else:
        bits = 32
    mask_i = resolv_nbo_i(mask)
    cidr = None
    for i in range(0, bits + 1):
        if ((1 << i) - 1) << (bits - i) == mask_i:
            cidr = i
            break
    return cidr
    
def addr_ctoa(cidr):
    if not (0 <= cidr <= 32):
        return None
        
    return addr_itoa(((1 << cidr) - 1) << 32 - cidr)
    
def compress_address(addr):
    if not is_ipv6(addr):
        return addr

    _addr = copy.copy(addr)

    while True:
        addr = re.sub(r'\A0:0:0:0:0:0:0:0\Z', '::', _addr)
        if (addr != _addr):
            break
        addr = re.sub(r'\b0:0:0:0:0:0:0\b', ':', _addr)
        if (addr != _addr):
            break
        addr = re.sub(r'\b0:0:0:0:0:0\b', ':', _addr)
        if (addr != _addr):
            break
        addr = re.sub(r'\b0:0:0:0:0\b', ':', _addr)
        if (addr != _addr):
            break
        addr = re.sub(r'\b0:0:0:0\b', ':', _addr)
        if (addr != _addr):
            break
        addr = re.sub(r'\b0:0:0\b', ':', _addr)
        if (addr != _addr):
            break
        addr = re.sub(r'\b0:0\b', ':', _addr)
        if (addr != _addr):
            break
        break
    return re.sub(r':{3,}', '::', addr)
    
def addr_ntoa(addr):
    """Converts a NBO address to an ASCII string (with IPv6 support)"""
    if len(addr) == 4:
        try:
            retval = '.'.join(str(oct) for oct in struct.unpack('4B', bytes(addr)))
        except TypeError:
            retval = '.'.join(str(oct) for oct in struct.unpack('4B', bytes(addr, 'latin-1')))
        return retval
        
    if len(addr) == 16:
        try:
            retval = ':'.join(('%x' % int(i)) for i in struct.unpack('>8H', bytes(addr)))
        except TypeError:
            retval = ':'.join(('%x' % int(i)) for i in struct.unpack('>8H', bytes(addr, 'latin-1')))
        return compress_address(retval)
        
    raise RuntimeError("Invalid address format")
    
def addr_iton(addr, v6 = False):
    """Converts an integer to a NBO address"""
    if addr < 0x100000000 and v6 is False:
        return str(struct.pack('>L', addr).decode('latin-1'))
    else:
        w = []
        w[0] = (addr >> 96) & 0xffffffff
        w[1] = (addr >> 64) & 0xffffffff
        w[2] = (addr >> 32) & 0xffffffff
        w[3] = addr & 0xffffffff
        return str(struct.pack('>4L', *w).decode('latin-1'))
    
def addr_ntoi(addr):
    """Converts a NBO address to an integer"""
    try:
        bits = struct.unpack(">{}L".format(len(addr) / 4), bytes(addr))
    except TypeError:
        bits = struct.unpack(">{}L".format(len(addr) / 4), bytes(addr, 'latin-1'))
    
    if len(bits) == 1:
        return bits[0]
        
    if len(bits) == 4:
        val = 0
        for i in bits:
            val += (bits[i] << (96 - (i * 32)))
            
        return val
        
    raise RuntimeError("Invalid address format")

def eth_aton(mac):
    return str(struct.pack("6B", *map(lambda c: int(c, 16), str(mac).split(':'))).decode('latin-1'))

def eth_ntoa(binary):
    try: 
        return ':'.join(map(lambda x: "%.2x" % x, struct.unpack("6B", bytes(binary)))).upper()
    except TypeError:
        return ':'.join(map(lambda x: "%.2x" % x, struct.unpack("6B", bytes(binary, 'latin-1')))).upper()
    
def cidr_crack(cidr, v6 = False):
    tmp = cidr.split('/')
    
    try:
        tst, scope = tmp[0].split("%", 1)
    except ValueError:
        # "need more than 1 value to unpack"
        tst = tmp[0].split("%", 1)[0]
        scope = ""
        
    addr = addr_atoi(tst)
    
    bits = 32
    mask = 0
    use6 = False
    
    if (addr > 0xffffffff or v6 or re.search(r':', cidr) is not None):
        use6 = True
        bits = 128
        
    mask = (2 ** bits) - (2 ** (bits - int(tmp[1])))
    base = addr & mask
    
    stop = base + (2 ** (bits - int(tmp[1]))) - 1
    return [addr_itoa(base, use6) + scope, addr_itoa(stop, use6) + scope]
    
def net2bitmask(netmask):
    nmask = resolv_nbo(netmask)
    imask = addr_ntoi(nmask)
    bits = 32

    if imask > 0xFFFFFFFF:
        bits = 128

    for bit in range(0, bits):
        p = 2 ** bit
        if ((imask & p) == p):
            return (bits - bit)

    return 0

def bit2netmask(bitmask, ipv6 = False):
    if bitmask > 32 or ipv6:
        i = ((~((2 ** (128 - bitmask)) - 1)) & (2 ** 128 - 1))
        n = addr_iton(i, True)
        return addr_ntoa(n)
    else:
        tmp1 = str(struct.pack('>L', *[(~((2 ** (32 - bitmask)) - 1)) & 0xFFFFFFFF]).decode('latin-1'))
        try:
            octets = struct.unpack('4B', bytes(tmp1))
        except TypeError:
            octets = struct.unpack('4B', bytes(tmp1, 'latin-1'))
        octets = map(lambda octet: str(octet), octets)
        return '.'.join(octets)

class TcpSocketPairHelper(object):
    def __init__(self):
        self.lsock = None
        self.rsock = None
        self.laddr = '127.0.0.1'
        self.lport = 0
        self.threads = []
        self.mutex = threading.Lock()
        
    def server_thread(self):
        server = None
        
        with self.mutex:
            server = SocketServer.TCPServer((self.laddr, 0), None)
            caddr, self.lport = server.socket.getsockname()
            
            thred = threading.Thread(target = self.client_thread, name = "TcpSocketPairClient")
            thred.daemon = True
            self.threads.append(thred)
            thred.start()
            
        self.lsock, _ = server.socket.accept()
        server.server_close()
        
    def client_thread(self):
        with self.mutex:
            self.rsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
            self.rsock.connect((self.laddr, self.lport))
            
    def make_socket_pair(self):
        thred = threading.Thread(target = self.server_thread, name = "TcpSocketPairServer")
        thred.daemon = True
        self.threads.append(thred)
        thred.start()
        
        for t in self.threads:
            t.join()
            
        return [self.lsock, self.rsock]
        
def tcp_socket_pair():
    return TcpSocketPairHelper().make_socket_pair()
    
def udp_socket_pair():
    laddr = '127.0.0.1'
    lsock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM, proto = 0)
    lsock.bind((laddr, 0))
    
    rsock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM, proto = 0)
    rsock.bind((laddr, 0))
    
    raddr, rport = lsock.getsockname()
    rsock.connect((raddr, rport))
    
    saddr, sport = rsock.getsockname()
    lsock.connect((saddr, sport))
    
    return [lsock, rsock]

class SocketEvents(object):
    def __init__(self):
        pass
        
    def on_before_socket_create(self, comm, param):
        pass
        
    def on_socket_created(self, comm, sock, param):
        pass
        
class SocketComm(object):
    def __init__(self):
        self.handlers_rwlock = None
        self.handlers = None
        pass
        
    def create(self, param):
        raise NotImplementedError
        
    def register_event_handler(self, handler):
        try:
            if self.handlers is None:
                self.handlers = []
            
            self.handlers.append(handler)
        except Exception as e:
            pass
        
    def deregister_event_handler(self, handler):
        try:
            if self.handlers is not None:
                self.handlers.remove(handler)
        except Exception as e:
            pass
            
    def each_event_handler(self, func=None):
        try:
            if self.handlers is not None:
                if func is not None:
                    [func(handler) for handler in self.handlers]
                return self.handlers
        except Exception as e:
            pass
    
    def notify_before_socket_create(self, comm, param):
        try:
            if self.each_event_handler() is not None:
                for handler in self.each_event_handler():
                    handler.on_before_socket_create(comm, param)
        except Exception as e:
            pass
            
    def notify_socket_created(self, comm, sock, param):
        try:
            if self.each_event_handler() is not None:
                for handler in self.each_event_handler():
                    handler.on_socket_created(comm, sock, param)
        except Exception as e:
            pass


class SockBase(object):
    def __init__(self, *args, **kwargs):
        if 'sock' in kwargs:
            self.sock = kwargs['sock']
        else:
            raise RuntimeError("SockBase did not wrap socket object!")
        if 'sslsock' in kwargs:
            self.sslsock = kwargs['sslsock']
        else:
            self.sslsock = None
        if 'sslctx' in kwargs:
            self.sslctx = kwargs['sslctx']
        else:
            self.sslctx = None
        if hasattr(self, "context") == False:
            self.context = None

    def bind(self, *args, **kwargs):
        return self.sock.bind(*args, **kwargs)

    def close(self):
        return self.sock.close()

    def connect(self, *args, **kwargs):
        return self.sock.connect(*args, **kwargs)

    def connect_to(self, *args, **kwargs):
        return self.sock.connect_to(*args, **kwargs)

    @property
    def fd(self):
        return self

    def fileno(self):
        try:
            return self.sock.fileno()
        except socket.error as e:
            if (e.errno == socket.EBADF):
                return -1
        except Exception as e:
            return -1

    def getpeername(self):
        return self.sock.getpeername()
        
    def getsockname(self):
        return self.sock.getsockname()

    def getsockopt(self, *args, **kwargs):
        return self.sock.getsockopt(*args, **kwargs)

    def initsock(self, params = None):
        if params is not None:
            self.peerhost = params.peerhost
            self.peerport = params.peerport
            self.localhost = params.localhost
            self.localport = params.localport
            self.context = params.context or {}
            if params.v6 == True:
                self.ipv = 6
            else:
                self.ipv = 4
        
    def ioctl(self, *args, **kwargs):
        return self.sock.ioctl(*args, **kwargs)
        
    def listen(self, *args, **kwargs):
        return self.sock.listen(*args, **kwargs)
        
    def makefile(self, *args, **kwargs):
        return self.sock.makefile(*args, **kwargs)
        
    def recv(self, *args, **kwargs):
        return self.sock.recv(*args, **kwargs)
        
    def recvfrom(self, *args, **kwargs):
        return self.sock.recvfrom(*args, **kwargs)
        
    def recvfrom_into(self, *args, **kwargs):
        return self.sock.recvfrom_into(*args, **kwargs)
        
    def recv_into(self, *args, **kwargs):
        return self.sock.recv_into(*args, **kwargs)
        
    def send(self, *args, **kwargs):
        return self.sock.send(*args, **kwargs)
        
    def sendall(self, *args, **kwargs):
        return self.sock.sendall(*args, **kwargs)
        
    def sendto(self, *args, **kwargs):
        return self.sock.sendto(*args, **kwargs)
            
    def setblocking(self, *args, **kwargs):
        return self.sock.setblocking(*args, **kwargs)
        
    def settimeout(self, *args, **kwargs):
        return self.sock.settimeout(*args, **kwargs)
        
    def gettimeout(self):
        return self.sock.gettimeout()
        
    def setsockopt(self, *args, **kwargs):
        return self.sock.setsockopt(*args, **kwargs)
        
    def shutdown(self, *args, **kwargs):
        return self.sock.shutdown(*args, **kwargs)
        
    @property
    def family(self):
        return self.sock.family
        
    @property
    def type(self):
        return self.sock.type
        
    @property
    def proto(self):
        return self.sock.proto
        
class TcpSock(SockBase, amncore.mixins.StreamMixin):
    def __init__(self, *args, **kwargs):
        return super(TcpSock, self).__init__(*args, **kwargs)

    @classmethod
    def create(self, **kwargs):
        kwargs['Proto'] = 'tcp'
        return self.create_param(SocketParams.from_kwargs(**kwargs))

    @classmethod
    def create_param(self, param):
        param.proto = 'tcp'
        return create_param(param)

    def shutdown(how = socket.SHUT_RDWR):
        try:
            return super(TcpSock, self).shutdown(how)
        except Exception as e:
            pass

    @property
    def type(self):
        return 'tcp'
        
class UdpSock(SockBase):
    def __init__(self, *args, **kwargs):
        return super(UdpSock, self).__init__(*args, **kwargs)

    @classmethod
    def create(self, **kwargs):
        kwargs['Proto'] = 'udp'
        if 'LocalHost' in kwargs:
            kwargs['Server'] = True
        
        return self.create_param(SocketParams.from_kwargs(**kwargs))

    @classmethod
    def create_param(self, param):
        param.proto = 'udp'
        return create_param(param)

    def write(self, gram):
        try:
            return self.send(gram)
        except socket.error as e:
            if (e.errno == errno.EHOSTUNREACH or e.errno == errno.ENETDOWN or e.errno == errno.ENETUNREACH or e.errno == errno.ENETRESET or e.errno == errno.EHOSTDOWN or e.errno == errno.EACCES or e.errno == errno.EINVAL or e.errno == errno.EADDRNOTAVAIL):
                return None

    def put(self, gram):
        return self.write(gram)

    def read(self, length = 65535):
        if length < 0:
            length = 65535
        return self.recv(length)

    def timed_read(self, length = 65535, timeout = 10):
        try:
            rv = select.select([self.fd], [], [], timeout)
            if (rv is not None) and (rv != ([], [], [])) and (len(rv[0]) > 0) and (rv[0][0] == self.fd):
                return self.read(length)
            else:
                return ''
        except Exception as e:
            return ''

    def sendto(self, gram, peerhost, peerport, flags = 0):
        peer = resolv_nbo(peerhost)
        if (len(peer) == 4 and self.ipv == 6):
            peerhost = getaddress(peerhost, True)
            if peerhost[0:7].lower() != '::ffff:':
                peerhost = '::ffff:' + peerhost

        try:
            return self.sock.sendto(gram, flags, getsockaddr(peerhost, peerport))
        except socket.error as e:
            if (e.errno == errno.EHOSTUNREACH or e.errno == errno.ENETDOWN or e.errno == errno.ENETUNREACH or e.errno == errno.ENETRESET or e.errno == errno.EHOSTDOWN or e.errno == errno.EACCES or e.errno == errno.EINVAL or e.errno == errno.EADDRNOTAVAIL):
                return None

    def recvfrom(self, length = 65535, timeout = 10):
        try:
            rv = select.select([self.fd], [], [], timeout)
            if (rv is not None) and (rv != ([], [], [])) and (len(rv[0]) > 0) and (rv[0][0] == self.fd):
                self.sock.setblocking(0)
                data, saddr = self.sock.recvfrom(length)
                af, host, port = parsesockaddr(saddr)
                
                return [data, host, port]
            else:
                return ['', None, None]
        except socket.error as e:
            return ['', None, None]
        except bombfuse.TimeoutError:
            return ['', None, None]
        except Exception as e:
            return ['', None, None]
            
    def get(self, timeout=None):
        data, saddr, sport = self.recvfrom(65535, timeout)
        return data

    @property
    def type(self):
        return 'udp'
        
class IpSock(SockBase):
    def __init__(self, *args, **kwargs):
        return super(IpSock, self).__init__(*args, **kwargs)

    @classmethod
    def create(self, **kwargs):
        kwargs['Proto'] = 'ip'
        return self.create_param(SocketParams.from_kwargs(**kwargs))

    @classmethod
    def create_param(self, param):
        param.proto = 'ip'
        return create_param(param)

    def write(self, gram):
        raise RuntimeError("IP sockets must use sendto(), not write()")
        
    def read(self, length = 65535):
        raise RuntimeError("IP sockets must use recvfrom(), not read()")
        
    def sendto(self, gram, peerhost, flags = 0):
        dest = getsockaddr(peerhost, 0)
        
        if (amncore.compat.is_freebsd() == True or amncore.compat.is_netbsd() == True or amncore.compat.is_macosx() == True):
            # byteswap
            try:
                tmp_unpakt = struct.unpack(">H", bytes(gram[2:4]))
            except TypeError:
                tmp_unpakt = struct.unpack(">H", bytes(gram[2:4], 'latin-1'))
            tmp_pakt = str(struct.pack("@h", *tmp_unpakt).decode('latin-1'))
            
            try:
                tmp_unpakt2 = struct.unpack(">H", bytes(gram[6:8]))
            except TypeError:
                tmp_unpakt2 = struct.unpack(">H", bytes(gram[6:8], 'latin-1'))
            tmp_pakt2 = str(struct.pack("@h", *tmp_unpakt2).decode('latin-1'))
            
            gram[2] = tmp_pakt[0]
            gram[3] = tmp_pakt[1]
            gram[6] = tmp_pakt2[0]
            gram[7] = tmp_pakt2[1]

        try:
            return self.sock.sendto(gram, flags, dest)
        except socket.error as e:
            return None
            
    def recvfrom(self, length = 65535, timeout = 10):
        try:
            rv = select.select([self.sock], [], [], timeout)
            if rv != ([],[],[]) and rv[0][0] == self.sock:
                data, saddr = super(IpSock, self).recvfrom(length)
                af, host, port = parsesockaddr(saddr)
                
                return [data, host]
            else:
                return ['', None]
        except Exception as e:
            return ['', None]
            
    def get(self, timeout = None):
        data, saddr = self.recvfrom(65535, timeout)
        return data

    def put(self, gram):
        return self.write(gram)

    @property
    def type(self):
        return 'ip'
        
class TcpServer(SockBase, amncore.mixins.StreamServerMixin):
    def __init__(self, *args, **kwargs):
        return super(TcpServer, self).__init__(*args, **kwargs)
    
    @classmethod
    def create(self, **kwargs):
        kwargs['Proto'] = 'tcp'
        kwargs['Server'] = True
        return self.create_param(SocketParams.from_kwargs(**kwargs))

    @classmethod
    def create_param(self, param):
        param.proto = 'tcp'
        param.server = True
        return create_param(param)

    def accept(self, **kwargs):
        try:
            r, _, _ = select.select([self.fd], [], [], 0.5)
            if (r is not None) and (r != []) and (self.fd in r):
                _sock, addr = self.sock.accept()
            else:
                return None
        except socket.error as e:
            return None

        retval = None
        if _sock is not None:
            retval = TcpSock(sock = _sock)
            if addr is not None:
                retval.peerhost = addr[0]
                retval.peerport = addr[1]

        return retval

class LocalComm(SocketComm):
    def __init__(self):
        return super(SocketComm, self).__init__()

    def create(self, param):
        if param.proto == 'tcp':
            return self.create_by_type(param, socket.SOCK_STREAM, socket.IPPROTO_TCP)
        elif param.proto == 'udp':
            return self.create_by_type(param, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        elif param.proto == 'ip':
            return self.create_ip(param)
        else:
            raise NotImplementedError("Unknown protocol: {}".format(param.proto))

    def create_ip(self, param):
        self.notify_before_socket_create(self, param)
        sock = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_RAW)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_HDRINCL, 1)

        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        if param.bare == False:
            # wrap socket in IpSock class object
            _sock = sock
            sock = IpSock(sock = _sock)
            sock.initsock(param)

        self.notify_socket_created(self, sock, param)

        return sock
    
    def create_by_type(self, param, type, proto = 0):
        if socket.has_ipv6 == True:
            if param.localhost is not None:
                local = resolv_nbo(param.localhost)
            else:
                local = None

            if param.peerhost is not None:
                peer = resolv_nbo(param.peerhost)
            else:
                peer = None

            if type == socket.SOCK_DGRAM and amncore.compat.is_linux() == True and local is None and peer is None:
                param.v6 = True
            elif (local is not None and len(local) == 16) or (peer is not None and len(peer) == 16):
                param.v6 = True

            if param.v6 == True:
                if local is not None and len(local) == 4:
                    if local == "\x00\x00\x00\x00":
                        param.localhost = '::'
                    elif local == "\x7f\x00\x00\x01":
                        param.localhost = "::1"
                    else:
                        param.localhost = "::ffff:" + getaddress(param.localhost, True)
                if peer is not None and len(peer) == 4:
                    if peer == "\x00\x00\x00\x00":
                        param.peerhost = '::'
                    elif peer == "\x7f\x00\x00\x01":
                        param.peerhost = "::1"
                    else:
                        param.peerhost = "::ffff:" + getaddress(param.peerhost, True)
        else:
            param.v6 = False

        self.notify_before_socket_create(self, param)

        sock = None
        if param.v6 == True:
            sock = socket.socket(socket.AF_INET6, type, proto)
        else:
            sock = socket.socket(socket.AF_INET, type, proto)
        bare_sock = sock

        if ((param.localport is not None) and (param.localport != 0)) or (param.localhost is not None):
            try:
                if amncore.compat.is_windows() == False:
                    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.bind(getsockaddr(param.localhost, param.localport))
            except socket.error as e:
                if e.errno == errno.EADDRNOTAVAIL or e.errno == errno.EADDRINUSE:
                    sock.close()
                    raise amncore.xcepts.BindError(param.localhost, param.localport)

        if type == socket.SOCK_DGRAM:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        if param.server == True:
            sock.listen(256)
            
            if param.bare == False:
                cls = TcpServer
                if param.ssl == True:
                    try:
                        import securesocket
                    except ImportError:
                        from . import securesocket
                    cls = securesocket.SslTcpServer
                sock = cls(sock = bare_sock)
                sock.initsock(param)
        else:
            chain = []

            if param.peerhost is not None:
                ip = param.peerhost
                port = param.peerport

                if param.proxies is not None:
                    chain = copy.copy(param.proxies)
                    chain.append(['host', param.peerhost, param.peerport])
                    ip = chain[0][1]
                    port = int(chain[0][2])

                try:
                    sock.settimeout(param.timeout)
                    sock.connect(getsockaddr(ip, port))
                except socket.timeout as e:
                    sock.close()
                    raise amncore.xcepts.ConnectionTimeout(ip, port)
                except socket.error as e:
                    if (e.errno == errno.EHOSTUNREACH or e.errno == errno.ENETDOWN or e.errno == errno.ENETUNREACH or e.errno == errno.ENETRESET or e.errno == errno.EHOSTDOWN or e.errno == errno.EACCES or e.errno == errno.EINVAL or e.errno == errno.ENOPROTOOPT):
                        sock.close()
                        raise amncore.xcepts.HostUnreachable(ip, port)
                    elif (e.errno == errno.EADDRNOTAVAIL or e.errno == errno.EADDRINUSE):
                        sock.close()
                        raise amncore.xcepts.InvalidDestination(ip, port)
                    elif (e.errno == errno.ETIMEDOUT or e.message == 'timed out'):
                        sock.close()
                        raise amncore.xcepts.ConnectionTimeout(ip, port)
                    elif (e.errno == errno.ECONNRESET or e.errno == errno.ECONNREFUSED or e.errno == errno.ENOTCONN or e.errno == errno.ECONNABORTED):
                        sock.close()
                        raise amncore.xcepts.ConnectionRefused(ip, int(port))
                finally:
                    sock.settimeout(None)
            
            if param.bare == False:
                if param.proto == 'tcp':
                    cls = TcpSock
                    sock = cls(sock = bare_sock)
                    sock.initsock(param)
                elif param.proto == 'udp':
                    cls = UdpSock
                    sock = cls(sock = bare_sock)
                    sock.initsock(param)

            if len(chain) > 1:
                for i, proxy in enumerate(chain):
                    try:
                        next_hop = chain[i + 1]
                    except Exception as e:
                        next_hop = None
                    if next_hop is not None:
                        self.proxy(sock, proxy[0], next_hop[1], next_hop[2])
            
            if (param.bare == False and param.ssl == True):
                try:
                    import securesocket
                except ImportError:
                    from . import securesocket
                cls = securesocket.SslTcpSock
                sock = cls(sock = bare_sock)
                sock.initsock(param)

        self.notify_socket_created(self, sock, param)
        return sock

    def proxy(self, sock, type, host, port):
        if type.lower() == 'sapni':
            packet_type = 'NI_ROUTE'
            route_info_ver = 2
            ni_ver = 39
            num_entries = 2
            talk_mode = 1
            num_rest_nodes = 1
            
            shost, sport = sock.getpeername()
            first_route_item = shost + str(struct.pack("B", 0).decode('latin-1')) + str(sport) + str(struct.pack("bb", 0, 0).decode('latin-1'))
            route_data = str(struct.pack(">L", len(first_route_item)).decode('latin-1')) + first_route_item
            route_data += (host + str(struct.pack("B", 0).decode('latin-1')) + str(port) + str(struct.pack("bb", 0, 0).decode('latin-1')))
            
            ni_packet = packet_type + str(struct.pack("8b", 0, route_info_ver, ni_ver, num_entries, talk_mode, 0, 0, num_rest_nodes).decode('latin-1'))
            
            ni_packet += (str(struct.pack(">L", len(route_data) - 4).decode('latin-1')) + route_data)
            
            ni_packet = (str(struct.pack(">L", len(ni_packet)).decode('latin-1')) + ni_packet)
            
            size = sock.put(ni_packet)
            
            if size != len(ni_packet):
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Failed to send the entire packet to the proxy server")
                
            try:
                ret_len = struct.unpack(">L", bytes(sock.get_once(4, 30)))[0]
                if ret_len and ret_len != 0:
                    ret = sock.get_once(ret_len, 30)
            except TypeError:
                ret_len = struct.unpack(">L", bytes(sock.get_once(4, 30), 'latin-1'))[0]
                if ret_len and ret_len != 0:
                    ret = sock.get_once(ret_len, 30)
            except IOError:
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Failed to receive a response from the proxy server")
            except struct.error:
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Failed to parse a response from the proxy server")
                
            if ret and len(ret) < 4:
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Failed to receive a complete response from the proxy server")
                
            if re.search(r'NI_RTERR', ret) is not None:
                if re.search(r'timed out', ret) is not None:
                    raise amncore.xcepts.ConnectionProxyError(host, port, type, "Connection to remote host {} timed out".format(host))
                elif re.search(r'refused', ret) is not None:
                    raise amncore.xcepts.ConnectionProxyError(host, port, type, "Connection to remote port {} refused".format(port))
                elif re.search(r'denied', ret) is not None:
                    raise amncore.xcepts.ConnectionProxyError(host, port, type, "Connection to {}:{} blocked by ACL".format(host, port))
                else:
                    raise amncore.xcepts.ConnectionProxyError(host, port, type, "Connection to {}:{} failed (unknown error)".format(host, port))
            elif re.search(r'NI_PONG', ret) is not None:
                pass
            else:
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Connection to {}:{} failed (unknown error)".format(host, port))
        elif type.lower() == 'http':
            setup = "CONNECT {}:{} HTTP/1.0\r\n\r\n".format(host, port)
            size = sock.put(setup)
            if size != len(setup):
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Failed to send the entire packet to the proxy server")
                
            try:
                ret = sock.get_once(39, 30)
            except IOError:
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Failed to receive a response from the proxy server")
                
            if ret is None:
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Failed to receive a response from the proxy server")
                
            resp = http_packet.HttpResponse()
            resp.update_cmd_parts(re.split(r'\r?\n', ret)[0])
            
            if resp.code != 200:
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Proxy server responded with non-OK code ({} {})".format(resp.code, resp.message))
        elif type.lower() == 'socks4':
            setup = str(struct.pack(">BBH", 4, 1, int(port)).decode('latin-1')) + addr_aton(host) + paynspray.rando.rand_text_alpha(random.randrange(8) + 1) + "\x00"
            size = sock.put(setup)
            if size != len(setup):
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Failed to send the entire packet to the proxy server")
            
            try:
                ret = sock.get_once(8, 30)
            except IOError:
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Failed to receive a response from the proxy server")
                
            if ret is None or len(ret) < 8:
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Failed to receive a complete response from the proxy server")
                
            if ret[1:2] != '\x5a':
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Proxy server responded with error code {}".format(struct.unpack("B", str(bytes(ret[0:1], 'latin-1')))[0]))
        elif type.lower() == 'socks5':
            auth_methods = str(struct.pack("BBB", 5, 1, 0).decode('latin-1'))
            size = sock.put(auth_methods)
            if size != len(auth_methods):
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Failed to send the entire packet to the proxy server")
                
            ret = sock.get_once(2, 30)
            if ret is not None and ret[1:2] == "\xff":
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "The proxy server requires authentication")
            elif ret is None:
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Failed to receive a response from the proxy server")
            
            if is_ipv4(host):
                addr = socket.inet_pton(socket.AF_INET, host)
                setup = str(struct.pack("4B", 5, 1, 0, 1).decode('latin-1')) + addr + str(struct.pack(">H", int(port)).decode('latin-1'))
            elif supports_ipv6() == True and is_ipv6(host):
                addr = socket.inet_pton(socket.AF_INET6, host)
                setup = str(struct.pack("4B", 5, 1, 0, 4).decode('latin-1')) + addr + str(struct.pack(">H", int(port)).decode('latin-1'))
            else:
                setup = str(struct.pack("4B", 5, 1, 0, 3).decode('latin-1')) + str(struct.pack("B", len(host)).decode('latin-1')) + host + str(struct.pack(">H", int(port)).decode('latin-1'))
                
            size = sock.put(setup)
            if size != len(setup):
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Failed to send the entire packet to the proxy server")
                
            try:
                response = sock.get_once(10, 30)
            except IOError:
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Failed to receive a response from the proxy server")
                
            if response is None or len(response) < 10:
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Failed to receive a complete response from the proxy server")
            
            if response[1:2] != "\x00":
                raise amncore.xcepts.ConnectionProxyError(host, port, type, "Proxy server responded with error code {}".format(struct.unpack("B", str(bytes(ret[1:2], 'latin-1')))[0]))
        else:
            raise ValueError("Invalid proxy type '{}' specified".format(type))
            
            




import copy
import enum
try:
    import HTMLParser
except ImportError:
    import html.parser as HTMLParser
import json
import paynspray.compression
import paynspray.encoding
import paynspray.rando
import random
import re
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse
from bs4 import BeautifulSoup
from collections import Mapping

DefaultProtocol = '1.1'

class HttpPacketHeader(Mapping):
    def __init__(self):
        self._dict = {}
        self.dcase_hash = {}
        self.junk_headers = None
        self.cmd_string = None
        self.fold = None
        
        self.reset()

    def __iter__(self):
        return iter(self._dict)
        
    def __len__(self):
        return len(self._dict)
        
    def __getitem__(self, key):
        try:
            rv = self._dict[key]
        except KeyError:
            rv = None
            
        if rv is None:
            try:
                rv = self.dcase_hash[str(key).lower()]
            except KeyError:
                rv = None
                
        return rv

    def __setitem__(self, key, value):
        stored = False
        for k in self._dict.keys():
            if (str(k).lower() == str(key).lower()):
                self._dict.update({k: value})
                stored = True
                
        if (stored == False):
            self._dict.update({key: value})
            
        self.dcase_hash.update({key.lower(): value})

    def __contains__(self, key):
        return self.__getitem__(key) is not None

    def __str__(self, prefix = ''):
        string = prefix
        
        if self.junk_headers and self.junk_headers == True:
            while len(string) < 4096:
                if self.fold and self.fold == True:
                    string += "X-{}:\r\n\t{}\r\n".format(paynspray.rando.rand_text_alphanumeric(random.randrange(30) + 5), paynspray.rando.rand_text_alphanumeric(random.randrange(1024) + 1))
                else:
                    string += "X-{}: {}\r\n".format(paynspray.rando.rand_text_alphanumeric(random.randrange(30) + 5), paynspray.rando.rand_text_alphanumeric(random.randrange(1024) + 1))
        
        if hasattr(self._dict, "iteritems") == True:
            for var, val in self._dict.iteritems():
                if self.fold and self.fold == True:
                    string += "{}:\r\n\t{}\r\n".format(var, val)
                else:
                    string += "{}: {}\r\n".format(var, val)
        else:
            for var, val in iter(self._dict.items()):
                if self.fold and self.fold == True:
                    string += "{}:\r\n\t{}\r\n".format(var, val)
                else:
                    string += "{}: {}\r\n".format(var, val)
                
        string += "\r\n"
        
        return string

    def from_str(self, header):
        self.reset()
        
        if (len(header) > 0 and re.search(r'\r\n', header) is None):
            header += "\r\n"
        
        header = re.sub(r'([^\r])\n', '\1' + "\r\n", header)
        
        header = re.sub(r':\s*\r\n\s+', ': ', header, flags = re.M | re.I)
        
        serch = re.search(r'.+\r\n', header)
        if (serch is not None):
            self.cmd_string = serch.group(0)
            header.replace(self.cmd_string, "", 1)
        
            
        for string in re.split(r'\r\n', header, flags = re.M):
            md = re.search(r'^(.+?)\s*:\s*(.+?)\s*$', string)
            if (md is not None):
                if md.group(1) in self._dict:
                    self._dict.update({md.group(1):self._dict[md.group(1)] + ", " + md.group(2)})
                else:
                    self._dict.update({md.group(1):md.group(2)})

    def from_list(self, lst):
        for e in lst:
            self._dict.update({e[0]:e[1]})
        return self

    def reset(self):
        self.cmd_string = ''
        self._dict.clear()
        self.dcase_hash.clear()
        return self

class ParseCode(enum.Enum):
    Completed = 1
    Partial = 2
    Error = 3

class ParseState(enum.Enum):
    ProcessingHeader = 1
    ProcessingBody = 2
    Completed = 3

class HttpPacket(object):
    def __init__(self):
        self.headers = HttpPacketHeader()
        self.auto_cl = True
        self.chunk_max_size = 1
        self.chunk_min_size = 1000
        self.body_bytes_left = 0
        self.compress = None
        self.error = None
        self.state = ParseState.ProcessingHeader
        self.incomplete = None
        self.transfer_chunked = False
        self.inside_chunk = False
        self._body = ''
        self.bufq = ''
        self.max_data = None

        self.reset()

    def __getitem__(self, key):
        if key in self.headers:
            return self.headers[key]
            
        if hasattr(self.headers, "iteritems") == True:
            for k, v in self.headers.iteritems():
                if (str(k).lower() == str(key).lower()):
                    return v
        else:
            for k, v in iter(self.headers.items()):
                if (str(k).lower() == str(key).lower()):
                    return v
        
        return None
        
    def __setitem__(self, key, value):
        self.headers[key] = value
    
    def __contains__(self, key):
        return key in self.headers

    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, string):
        self._body = string

    def parse(self, buf):
        self.bufq += str(buf)
        try:
            if (self.state == ParseState.ProcessingHeader):
                self.parse_header()
            
            if (self.state == ParseState.ProcessingBody):
                if (self.body_bytes_left == 0 and not self.transfer_chunked):
                    self.state = ParseState.Completed
                else:
                    self.parse_body()
        except Exception as e:
            self.error = e
            return ParseCode.Error

        if (self.state == ParseState.Completed):
            return ParseCode.Completed
        else:
            return ParseCode.Partial

    def reset(self):
        self.state = ParseState.ProcessingHeader
        self.transfer_chunked = False
        self.inside_chunk = False
        self.headers.reset()
        self.bufq = ''
        self.body = ''

    def reset_except_queue(self):
        self.state = ParseState.ProcessingHeader
        self.transfer_chunked = False
        self.inside_chunk = False
        self.headers.reset()
        self.body = ''

    @property
    def completed(self):
        if self.state == ParseState.Completed:
            return True

        if (self.state == ParseState.ProcessingBody and self.body_bytes_left < 0):
            return True

        return False

    def chunk(self, string, min_size = 1, max_size = 1000):
        chunked = ''
        
        if (min_size < 1):
            min_size = 1
            
        if (max_size < min_size):
            max_size = min_size
            
        while (len(string) > 0):
            chunk = string[0: random.randrange(max_size - min_size) + min_size]
            string = string[len(chunk):]
            chunked += (("%x" % len(chunk)) + "\r\n" + chunk + "\r\n")
        
        chunked += "0\r\n\r\n"
        return chunked

    def __str__(self):
        if self.body is None:
            content = None
        else:
            content = copy.copy(str(self.body))
        
        if content is not None:
            if self.compress is not None:
                if self.compress == 'gzip':
                    self.headers['Content-Encoding'] = 'gzip'
                    content = paynspray.compression.gzip(content)
                elif self.compress == 'deflate':
                    self.headers['Content-Encoding'] = 'deflate'
                    content = paynspray.compression.zlib_deflate(content)
                elif self.compress == 'none':
                    pass
                else:
                    raise RuntimeError("Invalid Content-Encoding")
                    
            if (self.auto_cl == True and self.transfer_chunked == True):
                raise RuntimeError("'Content-Length' and 'Transfer-Encoding: chunked' are incompatible")
            elif (self.auto_cl == True):
                self.headers['Content-Length'] = len(content)
            elif (self.transfer_chunked == True):
                if self.proto != '1.1':
                    raise RuntimeError("Chunked encoding is only available via HTTP 1.1")
                self.headers['Transfer-Encoding'] = 'chunked'
                content = self.chunk(content, self.chunk_min_size, self.chunk_max_size)
        
        string = self.headers.__str__(self.cmd_string)
        if content is not None:
            string += content
            
        return string

    def from_str(self, string):
        self.reset()
        return self.parse(string)

    @property            
    def cmd_string(self):
        try:
            return self.headers.cmd_string
        except AttributeError:
            return None

    def update_cmd_parts(self, string):
        pass

    def parse_header(self):
        head, data = re.split(r'\r?\n\r?\n', self.bufq, 1)
        
        if data is None:
            return
            
        self.headers.from_str(head)
        if data is not None:
            self.bufq = data
        else:
            self.bufq = ""
            
        self.body_bytes_left = -1
        if 'Content-Length' in self.headers and self.headers['Content-Length'] is not None:
            self.body_bytes_left = int(self.headers['Content-Length'])
            
        if ('Transfer-Encoding' in self.headers and str(self.headers['Transfer-Encoding']).lower() == 'chunked'):
            self.transfer_chunked = True
            self.auto_cl = False
            
        if (self.body_bytes_left == -1):
            if (self.transfer_chunked == False):
                if ('Connection' in self.headers and 'keep-alive' in str(self.headers['Connection']).lower()):
                    self.body_bytes_left = len(self.bufq)
                elif ('Content-Length' not in self.headers and re.search(r'HttpRequest', str(self.__class__)) is not None):
                    self.body_bytes_left = 0
        
        if self.headers.cmd_string is None:
            raise RuntimeError("Invalid command string")
            
        self.state = ParseState.ProcessingBody
        self.update_cmd_parts(self.headers.cmd_string)

    def parse_body(self):
        if len(self.bufq) == 0:
            return
            
        if self.transfer_chunked == True and self.inside_chunk != 1 and len(self.bufq) != 0:
            self.bufq = self.bufq.lstrip()
            
            if '\n' not in self.bufq:
                return
                
            clen = None
            serch = re.search(r'^[a-fA-F0-9]+\r?\n', self.bufq)
            if serch is not None:
                clen = serch.group(0)
                self.bufq.replace(clen, "", 1)
                
            if clen is not None:
                clen = clen.rstrip()
                
            if clen is None and len(self.bufq) == 0:
                return
                
            if clen is None:
                self.state = ParseState.Completed
                return
                
            self.body_bytes_left = int(clen, 16)
            
            if (self.body_bytes_left == 0):
                self.bufq = re.sub(r'^\r?\n(?s)', '', self.bufq)
                self.state = ParseState.Completed
                self.check_100()
                return
                
            self.inside_chunk = 1
        
        if self.body_bytes_left > 0:
            part = self.bufq[0:self.body_bytes_left]
            self.bufq = self.bufq[len(part):]
            self.body = self.body + part
            self.body_bytes_left -= len(part)
        else:
            self.body = self.body + self.bufq
            self.bufq = ''
            
        if (self.transfer_chunked == True and self.body_bytes_left == 0):
            self.inside_chunk = 0
            self.parse_body()
            return
            
        if (self.transfer_chunked == False and self.body_bytes_left == 0):
            self.state = ParseState.Completed
            self.check_100()
            return

    def check_100(self):
        pass

class HttpRequest(HttpPacket):

    PostRequests = ['POST', 'SEARCH']

    def __init__(self, method = 'GET', uri = '/', proto = DefaultProtocol):
        super(HttpRequest, self).__init__()
        self.method = method
        self.raw_uri = uri
        self.uri_parts = {}
        if proto is not None:
            self.proto = proto
        else:
            self.proto = DefaultProtocol
        self.chunk_min_size = 1
        self.chunk_max_size = 10
        self.uri_encode_mode = 'hex-normal'
        self.relative_resource = None
        self.junk_directories = None
        self.junk_slashes = None
        self.junk_self_referring_directories = None
        self.junk_params = None
        self.junk_end_of_url = None
        self.junk_pipeline = None
        self.junk_param_start = None

        self.update_uri_parts()

    def update_cmd_parts(self, string):
        parser = HTMLParser.HTMLParser()
        md = re.search(r'^(.+?)\s+(.+?)\s+HTTP\/(.+?)\r?\n?$', string)
        if md is not None:
            self.method = md.group(1)
            self.raw_uri = parser.unescape(md.group(2))
            self.proto = md.group(3)

            return self.update_uri_parts()
        else:
            raise RuntimeError("Invalid request command string")


    def update_uri_parts(self):
        if self.raw_uri is not None:
            mach = re.search(r'(.+?)\?(.*)$', self.raw_uri)
            if mach is not None:
                self.uri_parts.update({'QueryString':self.parse_cgi_qstring(mach.group(2))})
                self.uri_parts.update({'Resource':mach.group(1)})
            else:
                self.uri_parts.update({'QueryString':{}})
                self.uri_parts.update({'Resource':self.raw_uri})
                
        normalized = self.resource
        normalized = re.sub(r'(\/\.\/|\/\w+\/\.\.\/|\/\/)', '/', normalized)
        self.resource = normalized
        # self.uri_parts.update({'Resource':normalized})
        self.relative_resource = self.resource

    @property
    def uri(self):
        def junk_dirs_block():
            dirs = ''
            for i in range(0, random.randrange(5) + 5):
                dirs += '/' + paynspray.rando.rand_text_alpha(random.randrange(5) + 1) + '/..'
            return dirs + '/'

        if 'Resource' in self.uri_parts:
            string = copy.copy(self.uri_parts['Resource'])
        else:
            string = '/'

        if self.junk_self_referring_directories:
            string = re.sub(r'\/', '/.' * (random.randrange(3) + 1) + '/', string)
        
        if self.junk_param_start:
            string = re.sub(r'\/', '/%3f' + paynspray.rando.rand_text_alpha(random.randrange(5) + 1) + '=' + paynspray.rando.rand_text_alpha(random.randrange(10) + 1) + '/../', string)
        
        if self.junk_directories:
            string = re.sub(r'\/', self.junk_dirs_block(), string)

        if self.junk_slashes:
            string = re.sub(r'\/', '/' * (random.randrange(3) + 2), string)
            string = re.sub(r'^[\/]+', '/', string)

        if self.junk_end_of_url:
            string = re.sub(r'^\/', '/%20HTTP/1.0%0d%0a/../../', string)

        string = paynspray.encoding.uri_encode(string, self.uri_encode_mode)

        if self.method in self.PostRequests:
            if len(self.param_string) > 0:
                string += '?' + self.param_string

        return string
    
    @property
    def param_string(self):
        params = []
        
        if hasattr(self.uri_parts['QueryString'], "iteritems") == True:
            for param, value in self.uri_parts['QueryString'].iteritems():
                if self.junk_params:
                    for i in range(0, random.randrange(10) + 5):
                        params.append(paynspray.rando.rand_text_alpha(random.randrange(16) + 5) + '=' + paynspray.rando.rand_text_alpha(random.randrange(10) + 1))
                if type(value) == list:
                    for subvalue in value:
                        params.append(paynspray.encoding.uri_encode(param, self.uri_encode_mode) + '=' + paynspray.encoding.uri_encode(subvalue, self.uri_encode_mode))
                else:
                    if value is not None:
                        params.append(paynspray.encoding.uri_encode(param, self.uri_encode_mode) + '=' + paynspray.encoding.uri_encode(value, self.uri_encode_mode))
                    else:
                        params.append(paynspray.encoding.uri_encode(param, self.uri_encode_mode))
        else:
            for param, value in iter(self.uri_parts['QueryString'].items()):
                if self.junk_params:
                    for i in range(0, random.randrange(10) + 5):
                        params.append(paynspray.rando.rand_text_alpha(random.randrange(16) + 5) + '=' + paynspray.rando.rand_text_alpha(random.randrange(10) + 1))
                if type(value) == list:
                    for subvalue in value:
                        params.append(paynspray.encoding.uri_encode(param, self.uri_encode_mode) + '=' + paynspray.encoding.uri_encode(subvalue, self.uri_encode_mode))
                else:
                    if value is not None:
                        params.append(paynspray.encoding.uri_encode(param, self.uri_encode_mode) + '=' + paynspray.encoding.uri_encode(value, self.uri_encode_mode))
                    else:
                        params.append(paynspray.encoding.uri_encode(param, self.uri_encode_mode))

        if self.junk_params:
            for i in range(0, random.randrange(10) + 5):
                params.append(paynspray.rando.rand_text_alpha(random.randrange(32) + 5) + '=' + paynspray.rando.rand_text_alpha(random.randrange(64) + 5))

        return '&'.join(params)

    @uri.setter
    def uri(self, string):
        self.raw_uri = string
        self.update_uri_parts()

    def __str__(self):
        string = ''
        if self.junk_pipeline:
            host = ''
            if 'Host' in self.headers:
                host = "Host: {}\r\n".format(self.headers['Host'])
            string += "GET / HTTP/1.1\r\n{}Connection: Keep-Alive\r\n\r\n" * self.junk_pipeline
            self.headers['Connection'] = 'Closed'
        return string + super(HttpRequest, self).__str__()

    @property
    def body(self):
        string = super(HttpRequest, self).body
        if string is None:
            string = ''

        if len(string) > 0:
            return string

        if self.method in self.PostRequests:
            return self.param_string

        return ''

    @body.setter
    def body(self, string):
        self._body = string

    @property
    def cmd_string(self):
        if re.compile(r'^\d').match(self.proto) is not None:
            proto_str = "HTTP/{}".format(self.proto)
        else:
            proto_str = self.proto
    
        return "{} {} {}\r\n".format(self.method, self.uri, proto_str)

    @property
    def resource(self):
        try:
            return self.uri_parts['Resource']
        except KeyError as e:
            return None

    @resource.setter
    def resource(self, rsrc):
        self.uri_parts.update({'Resource':rsrc})

    @property
    def qstring(self):
        try:
            return self.uri_parts['QueryString']
        except KeyError as e:
            return None

    def parse_cgi_qstring(self, string):
        qstring = {}
        
        for vv in re.split(r'[;&]', string):
            var = vv
            val = ''
            
            mach = re.search(r'(.+?)=(.*)', string)
            if mach is not None:
                var = mach.group(1)
                val = mach.group(2)
                
            if var in qstring:
                if (type(qstring[val]) == list):
                    qstring[val].append(val)
                else:
                    curr = self.qstring[var]
                    qstring.update({var: [curr, val]})
            else:
                qstring.update({var: val})
                
        return qstring


class HttpGetRequest(HttpRequest):
    def __init__(self, uri = '/', proto = DefaultProtocol):
        super(HttpGetRequest, self).__init__("GET", uri, proto)

class HttpPostRequest(HttpRequest):
    def __init__(self, uri = '/', proto = DefaultProtocol):
        super(HttpPostRequest, self).__init__("POST", uri, proto)

class HttpPutRequest(HttpRequest):
    def __init__(self, uri = '/', proto = DefaultProtocol):
        super(HttpPutRequest, self).__init__("PUT", uri, proto)
        
class HttpResponse(HttpPacket):
    def __init__(self, code = 200, message = 'OK', proto = DefaultProtocol):
        super(HttpResponse, self).__init__()
        self.code = int(code)
        self.message = message
        self.proto = proto
        self.auto_cl = True
        self.chunk_min_size = 1
        self.chunk_max_size = 10
        self.count_100 = 0
        self.request = None

    def get_cookies(self):
        cookies = ""
        
        if 'Set-Cookie' in self.headers:
            set_cookies = self.headers['Set-Cookie']
            if set_cookies is None:
                return cookies

            mach = re.findall(r'\s?([^, ;]+?)=([^, ;]*?)[;,]', set_cookies)

            if mach is not None:
                key_vals = dict(mach)
                if hasattr(key_vals, "iteritems") == True:
                    for k, v in key_vals.iteritems():
                        name = str(k).lower()
                        if name == 'path':
                            continue
                        if name == 'expires':
                            continue
                        if name == 'domain':
                            continue
                        if name == 'max-age':
                            continue
                        cookies += "{}={}; ".format(k, v)
                else:
                    for k, v in iter(key_vals.items()):
                        name = str(k).lower()
                        if name == 'path':
                            continue
                        if name == 'expires':
                            continue
                        if name == 'domain':
                            continue
                        if name == 'max-age':
                            continue
                        cookies += "{}={}; ".format(k, v)
            else:
                return cookies
                
        return cookies.strip()
        
    def get_html_document(self):
        return BeautifulSoup(self.body, 'html.parser')
        
    def get_xml_document(self):
        return BeautifulSoup(self.body, 'xml')

    def get_json_document(self):
        _json = {}
        
        try:
            _json = json.loads(self.body)
        except Exception as e:
            pass

        return _json
        
    def get_html_meta_elements(self):
        bs = self.get_html_document()
        return bs.find_all("meta")

    def update_cmd_parts(self, string):
        mach = re.search(r'HTTP\/(.+?)\s+(\d+)\s?(.+?)\r?\n?$', string)
        if mach is not None:
            self.message = re.sub(r'\r', '', mach.group(3))
            self.code = int(mach.group(2))
            self.proto = mach.group(1)
        else:
            raise RuntimeError("Invalid response command string")

        self.check_100()

    def check_100(self):
        if self.code == 100 and (self.body_bytes_left == -1 or self.body_bytes_left == 0) and self.count_100 < 5:
            self.reset_except_queue()
            self.count_100 += 1

    @property
    def redirect(self):
        return self.code in [301, 302, 303, 307, 308]

    @property
    def cmd_string(self):
        if self.message is not None and len(self.message) > 0:
            return "HTTP/{} {}{}\r\n".format(self.proto, self.code, ' ' + self.message)
        else:
            return "HTTP/{} {}{}\r\n".format(self.proto, self.code, '')

    @property
    def redirection(self):
        try:
            return urlparse.urlsplit(self.headers['Location'])
        except Exception:
            return None

class HttpOKResponse(HttpResponse):
    def __init__(self, message = 'OK', proto = DefaultProtocol):
        super(HttpOKResponse, self).__init__(200, message, proto)

class HttpE404Response(HttpResponse):
    def __init__(self, message = 'Not Found', proto = DefaultProtocol):
        return super(HttpE404Response, self).__init__(404, message, proto)




import copy
import re
import socket
try:
    import sockets
except ImportError:
    from . import sockets

class SubnetWalker(object):
    def __init__(self, subnet, netmask):
        self.subnet = sockets.resolv_to_dotted(subnet)
        self.netmask = sockets.resolv_to_dotted(netmask)
        self.num_ips = None
        self.curr_ip = None
        self.curr_ip_idx = None
        
        self.reset()
        
    def reset(self):
        self.curr_ip = self.subnet.split('.')
        self.num_ips = (1 << (32 - int(sockets.net2bitmask(self.netmask))))
        self.curr_ip_idx = 0
        
    def next_ip(self):
        if (self.curr_ip_idx >= self.num_ips):
            return None
            
        if (self.curr_ip_idx > 0):
            self.curr_ip[3] = (int(self.curr_ip[3]) + 1) % 256
            if self.curr_ip[3] == 0:
                self.curr_ip[2] = (int(self.curr_ip[2]) + 1) % 256
            if self.curr_ip[2] == 0:
                self.curr_ip[1] = (int(self.curr_ip[1]) + 1) % 256
            if self.curr_ip[1] == 0:
                self.curr_ip[0] = (int(self.curr_ip[0]) + 1) % 256
                
        self.curr_ip_idx += 1
        
        return '.'.join(map(str, self.curr_ip))
            

class IpRange(object):
    def __init__(self, start = None, stop = None, **kwargs):
        self.start = start
        self.stop = stop
        self.options = kwargs
        
    def __eq__(self, other):
        return (other.start == self.start and other.stop == self.stop and other.is_ipv6 == self.is_ipv6 and other.options == self.options)
        
    def __len__(self):
        return (self.stop - self.start + 1)
        
    @property
    def is_ipv6(self):
        return ('ipv6' in self.options and self.options['ipv6'] == True)
        

class IpRangeWalker(object):
    def __init__(self, parse_obj):
        self.curr_range_index = 0
        self.curr_addr = None
        self.length = 0
        
        # copy constructor
        if isinstance(parse_obj, IpRangeWalker) == True:
            self.ranges = copy.copy(parse_obj.ranges)
        else:
            self.ranges = self.parse(parse_obj)
            
        self.reset()

    def __len__(self):
        return self.length
        
    def parse(self, parse_obj):
        if parse_obj is None:
            return None
        ranges = []
        for arg in [y for x in map(lambda a: a.split(' '), parse_obj.split(', ')) for y in x]:
            opts = {}
            
            v4range_mach = re.search(r'^([0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})-([0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})$', arg)
            if ":" in arg:
                addrs = arg.split('-', 1)
                
                if len(addrs) == 1:
                    try:
                        addr, scope_id = addrs[0].split('%')
                    except ValueError:
                        addr = addrs[0].split('%')[0]
                        scope_id = None
                        
                    if scope_id is not None:
                        opts['scope_id'] = scope_id
                        
                    opts['ipv6'] = True
                    
                    if sockets.is_ipv6(addr) == False:
                        return False
                    addr = sockets.addr_atoi(addr)
                    ranges.append(IpRange(addr, addr, **opts))
                    continue
                    
                try:
                    addr1, scope_id = addrs[0].split('%')
                except ValueError:
                    addr1 = addrs[0].split('%')[0]
                    scope_id = None
                    
                if scope_id is not None:
                    opts['scope_id'] = scope_id
                    
                try:
                    addr2, scope_id = addrs[0].split('%')
                except ValueError:
                    addr2 = addrs[0].split('%')[0]
                    scope_id = None
                    
                if scope_id is not None and 'scope_id' not in opts:
                    opts['scope_id'] = scope_id
                    
                if sockets.is_ipv6(addr1) == False or sockets.is_ipv6(addr2) == False:
                    return False
                    
                addr1 = sockets.addr_atoi(addr1)
                addr2 = sockets.addr_atoi(addr2)
                
                ranges.append(IpRange(addr1, addr2, **opts))
                continue

            elif '/' in arg:
                if re.search(r'[,-]', arg) is not None:
                    return False
                if len(re.findall("/", arg)) > 1:
                    return False
                
                try:
                    ip_part, mask_part = arg.split("/")
                except ValueError:
                    ip_part = arg.split("/")[0]
                    mask_part = None

                if ip_part is None or len(ip_part) == 0 or mask_part is None or len(mask_part) == 0:
                    return False

                if re.search(r'^[0-9]{1,2}$', mask_part) is None:
                    return False

                if int(mask_part) > 32:
                    return False

                if re.search(r'^\d{1,3}(\.\d{1,3}){1,3}$', ip_part) is not None:
                    if re.search(sockets.MATCH_IPV4, ip_part) is None:
                        return False

                try:
                    sockets.getaddress(ip_part)
                except (socket.gaierror, socket.error) as e:
                    return False

                expanded = self.expand_cidr(arg)
                if expanded is not None and not expanded == False:
                    ranges.append(expanded)
                else:
                    return False

            elif re.search(r'[^-0-9,.*]', arg) is not None:
                try:
                    ranges += map(lambda a: IpRange(a, a, **opts), sockets.resolv_nbo_i_list(arg))
                except (socket.gaierror, socket.error) as e:
                    return False

            elif v4range_mach is not None:
                try:
                    start, stop = sockets.addr_atoi(v4range_mach.group(1)), sockets.addr_atoi(v4range_mach.group(2))
                    if start > stop:
                        return False
                    ranges.append(IpRange(start, stop, **opts))
                except (socket.gaierror, socket.error) as e:
                    return False

            else:
                expanded = self.expand_nmap(arg)
                if expanded:
                    for r in expanded:
                        ranges.append(r)

        def list_uniq(duplicate): 
            final_list = [] 
            for num in duplicate: 
                if num not in final_list: 
                    final_list.append(num) 
            return final_list 

        ranges = list_uniq(ranges)
        return ranges
            
    def reset(self):
        if self.valid == False:
            return False
        self.curr_range_index = 0
        self.curr_addr = self.ranges[0].start
        self.length = 0
        for r in self.ranges:
            self.length += len(r)

        return self
    
    def next_ip(self):
        if not self.valid:
            return False

        if (self.curr_addr > self.ranges[self.curr_range_index].stop):
            try:
                if self.ranges[self.curr_range_index + 1] is None:
                    return False
            except IndexError:
                return False

            self.curr_range_index += 1

            self.curr_addr = self.ranges[self.curr_range_index].start
        addr = sockets.addr_itoa(self.curr_addr, self.ranges[self.curr_range_index].is_ipv6)

        if 'scope_id' in self.ranges[self.curr_range_index].options:
            addr = addr + '%' + self.ranges[self.curr_range_index].options['scope_id']

        self.curr_addr += 1
        return addr

    @property
    def valid(self):
        return ((self.ranges is not None or not self.ranges == False) and (len(self.ranges) != 0))
                    
    def expand_cidr(self, arg):
        try:
            start, stop = sockets.cidr_crack(arg)
        except ValueError:
            return False
            
        return IpRange(sockets.addr_atoi(start), sockets.addr_atoi(stop), ipv6 = (':' in arg))
        
    def expand_nmap(self, arg):
        if ':' in arg:
            return False
            
        if ',-' in arg or '-,' in arg:
            return False
            
        bytes = []
        sections = arg.split('.')
        if len(sections) != 4:
            return False
            
        for section in sections:
            if len(section) == 0:
                section = "0"
                
            if section == "*":
                section = "0-255"
            elif '*' in section:
                return False

            ranges = section.split(',', -1)
            sets = []
            for r in ranges:
                bounds = []
                if '-' in r:
                    bounds = r.split('-', -1)
                    if len(bounds) > 2:
                        return False
                
                    if bounds[0] is None or len(bounds[0]) == 0:
                        bounds[0] = 0
                        
                    if bounds[1] is None or len(bounds[0]) == 0:
                        bounds[1] = 255

                    bounds = map(lambda b: int(b), bounds)
                    if (bounds[0] > bounds[1]):
                        return False
                else:
                    bounds.insert(0, int(r))
                if bounds[0] > 255 or (len(bounds) > 1 and bounds[1] > 255):
                    return False
                if len(bounds) > 1 and (bounds[0] > bounds[1]):
                    return False
                if len(bounds) > 1:
                    for i in range(bounds[0], bounds[1] + 1):
                        sets.append(i)
                elif len(bounds) == 1:
                    sets.append(bounds[0])
            bytes.append(sorted(set(sets)))

        addrs = []
        for a in bytes[0]:
            for b in bytes[1]:
                for c in bytes[2]:
                    for d in bytes[3]:
                        ip = (a << 24) + (b << 16) + (c << 8) + d
                        addrs.append(ip)

        addrs = sorted(set(addrs))
        rng = IpRange()
        rng.options = {'ipv6':False}
        rng.start = addrs[0]

        ranges = []
        for idx in range(1, len(addrs)):
            if addrs[idx - 1] + 1 == addrs[idx]:
                continue
            else:
                rng.stop = addrs[idx - 1]
                ranges.append(copy.copy(rng))
                rng.start = addrs[idx]
        rng.stop = addrs[len(addrs) - 1]
        ranges.append(copy.copy(rng))
        return ranges
            
                
                


import amncore.sync
import amncore.xcepts
import bombfuse
import datetime
import OpenSSL
import random
import select
import socket
try:
    import sockets
except ImportError:
    from . import sockets
try:
    import x509
except ImportError:
    from . import x509

class SecureSock(object):

    class CertProv(object):
        @classmethod
        def ssl_generate_subject(self):
            import paynspray.rando
            nom = OpenSSL.crypto.X509Name(OpenSSL.crypto.X509().get_subject())

            st = str(paynspray.rando.rand_state())
            loc = str(paynspray.rando.rand_name()).capitalize()
            org = str(paynspray.rando.rand_name()).capitalize()
            cn = str(paynspray.rando.rand_hostname())

            nom.C = "US"
            nom.ST = st
            nom.L = loc
            nom.O = org
            nom.CN = cn
            return nom

        @classmethod
        def ssl_generate_issuer(self):
            import paynspray.rando
            nom = OpenSSL.crypto.X509Name(OpenSSL.crypto.X509().get_issuer())

            org = str(paynspray.rando.rand_name()).capitalize()
            cn = str(paynspray.rando.rand_name()).capitalize() + " " + str(paynspray.rando.rand_name()).capitalize()

            nom.C = "US"
            nom.O = org
            nom.CN = cn
            return nom
        
        @classmethod
        def ssl_generate_certificate(self):
            yr = 24 * 3600 * 365
            vf = datetime.datetime.now() - datetime.timedelta(seconds = (random.randrange(yr * 3) + yr))
            vt = vf + datetime.timedelta(seconds = ((random.randrange(9) + 1) * yr))
            vf = vf.strftime("%Y%m%d%H%M%SZ")
            vt = vt.strftime("%Y%m%d%H%M%SZ")
            subject = self.ssl_generate_subject()
            issuer = self.ssl_generate_issuer()
            key = OpenSSL.crypto.PKey()
            key.generate_key(OpenSSL.crypto.TYPE_RSA, 2048)
            cert = OpenSSL.crypto.X509()
            cert.set_version(2)
            cert.set_serial_number((random.randrange(0xFFFFFFFF) << 32) + random.randrange(0xFFFFFFFF))
            cert.set_subject(subject)
            cert.set_issuer(issuer)
            try:
                cert.set_notBefore(vf)
                cert.set_notAfter(vt)
            except TypeError:
                cert.set_notBefore(bytes(vf, 'ascii'))
                cert.set_notAfter(bytes(vt, 'ascii'))
            cert.set_pubkey(key)

        
            ef = OpenSSL.crypto.X509Extension
            cert.add_extensions([ef(b"basicConstraints", True, b"CA:FALSE")])
        
            cert.sign(key, 'sha256')
        
            return [key, cert, None]

    cert_prov = CertProv

    def __init__(self, *args, **kwargs):
        if hasattr(self, "sock") == False:
            self.sock = None
        return super(SecureSock, self).__init__(*args, **kwargs)

    @property
    def cert_provider(self):
        return self.cert_prov

    @cert_provider.setter
    def cert_provider(self, val):
        self.cert_prov = val

    @classmethod
    def ssl_parse_pem(self, ssl_cert):
        return x509.X509Cert.parse_pem(ssl_cert)

    @classmethod
    def ssl_generate_subject(self):
        return self.cert_prov.ssl_generate_subject()

    @classmethod
    def ssl_generate_issuer(self):
        return self.cert_prov.ssl_generate_issuer()

    @classmethod
    def ssl_generate_certificate(self):
        return self.cert_prov.ssl_generate_certificate()

    def makessl(self, params):
        import paynspray.rando
        if hasattr(params, "ssl_cert") == True and params.ssl_cert is not None:
            key, cert, chain = self.ssl_parse_pem(params.ssl_cert)
        else:
            key, cert, chain = self.ssl_generate_certificate()

        ctx = OpenSSL.SSL.Context(OpenSSL.SSL.TLSv1_2_METHOD)
        
        if key is not None:
            ctx.use_privatekey(key)
        if cert is not None:
            ctx.use_certificate(cert)
        if chain is not None and chain != []:
            ctx.add_extra_chain_cert(chain)

        ctx.set_options(0)
        
        if hasattr(params, "ssl_cipher") == True:
            if params.ssl_cipher is not None:
                ctx.set_cipher_list(params.ssl_cipher)

        if hasattr(OpenSSL.SSL, "OP_NO_COMPRESSION") == True:
            if hasattr(params, "ssl_compression") == True:
                if params.ssl_compression is not None and params.ssl_compression == False:
                    ctx.set_options(OpenSSL.SSL.OP_NO_COMPRESSION)

        ctx.set_session_id(paynspray.rando.rand_text(16))
        
        return ctx

    def allow_nonblock(self, sock = None):
        import amncore.compat
        if sock is None:
            if hasattr(self, "sock") == True:
                sock = self.sock
        avail = hasattr(sock, "setblocking")
        if avail == True and amncore.compat.is_windows() == True:
            return True
        return False
        
class SslTcpSock(sockets.TcpSock):
    def __init__(self, *args, **kwargs):
        if hasattr(self, "ssl_negotiated_version") == False:
            self.ssl_negotiated_version = None
        if hasattr(self, "sslsock") == False:
            self.sslsock = None
        if hasattr(self, "sslctx") == False:
            self.sslctx = None
        return super(SslTcpSock, self).__init__(*args, **kwargs)

    @classmethod
    def create(self, **kwargs):
        kwargs['SSL'] = True
        return self.create_param(sockets.SocketParams.from_kwargs(**kwargs))
    
    @classmethod
    def create_param(self, param):
        param.ssl = True
        return super(SslTcpSock, self).create_param(param)

    def initsock(self, params = None):
        super(SslTcpSock, self).initsock(params)

        SSLv2 = 1
        SSLv3 = 2
        SSLv23 = 3
        TLSv1 = 4
        TLSv1_1 = 5
        TLSv1_2 = 6

        version = OpenSSL.SSL.SSLv23_METHOD

        if params is not None and params.ssl_version is not None:
            if params.ssl_version == 'SSL2' or params.ssl_version == SSLv2:
                version = OpenSSL.SSL.SSLv2_METHOD
            elif params.ssl_version == 'SSL23' or params.ssl_version == SSLv23 or params.ssl_version == 'TLS' or params.ssl_version == 'Auto':
                version = OpenSSL.SSL.SSLv23_METHOD
            elif params.ssl_version == 'SSL3' or params.ssl_version == SSLv3:
                version = OpenSSL.SSL.SSLv3_METHOD
            elif params.ssl_version == 'TLS1' or params.ssl_version == 'TLS1.0' or params.ssl_version == TLSv1:
                version = OpenSSL.SSL.TLSv1_METHOD
            elif params.ssl_version == 'TLS1.1' or params.ssl_version == TLSv1_1:
                version = OpenSSL.SSL.TLSv1_1_METHOD
            elif params.ssl_version == 'TLS1.2' or params.ssl_version == TLSv1_2:
                version = OpenSSL.SSL.TLSv1_2_METHOD

        self.initsock_with_ssl_version(params, version)

        self.ssl_negotiated_version = version

        return self.ssl_negotiated_version

    def initsock_with_ssl_version(self, params, version):
        self.sslctx = OpenSSL.SSL.Context(version)

        if params is not None and params.ssl_client_cert is not None:
            try:
                tmp_cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, bytes(params.ssl_client_cert, 'utf-8'))
            except TypeError:
                tmp_cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, params.ssl_client_cert)
            self.sslctx.use_certificate(tmp_cert)
            
        if params is not None and params.ssl_client_key is not None:
            try:
                tmp_key = OpenSSL.crypto.load_privatekey(OpenSSL.crypto.FILETYPE_PEM, bytes(params.ssl_client_key, 'utf-8'))
            except TypeError:
                tmp_key = OpenSSL.crypto.load_privatekey(OpenSSL.crypto.FILETYPE_PEM, params.ssl_client_key)
            self.sslctx.use_privatekey(tmp_key)

        if params is not None and params.ssl_verify_mode is not None:
            verify_mode = getattr(OpenSSL.SSL, "VERIFY_{}".format(params.ssl_verify_mode))
        else:
            verify_mode = OpenSSL.SSL.VERIFY_PEER
            
        self.sslctx.set_verify(verify_mode, verify_cb)
        
        self.sslctx.set_options(OpenSSL.SSL.OP_ALL)

        if hasattr(params, "ssl_cipher") == True:
            if params.ssl_cipher is not None:
                try:
                    self.sslctx.set_cipher_list(bytes(params.ssl_cipher, 'ascii'))
                except TypeError:
                    self.sslctx.set_cipher_list(params.ssl_cipher)
            
        self.sslsock = OpenSSL.SSL.Connection(self.sslctx, self.sock)
        self.sslsock.set_connect_state()
        try:
            self.sslsock.set_tlsext_host_name(bytes(params.peerhost, 'ascii'))
        except TypeError:
            self.sslsock.set_tlsext_host_name(params.peerhost)
        
        def nonblock_handshake_proc():
            while True:
                try:
                    self.sslsock.do_handshake()
                    break
                except OpenSSL.SSL.WantReadError as e:
                    select.select([self.sslsock], [], [], 0.10)
                    continue
                except OpenSSL.SSL.WantWriteError as e:
                    select.select([], [self.sslsock], [], 0.10)
                    continue
                except Exception as e:
                    raise e

        if self.allow_nonblock() == False:
            try:
                self.sslsock.do_handshake()
            except socket.timeout as e:
                raise amncore.xcepts.ConnectionTimeout(params.peerhost, params.peerport)
        else:
            self.sslsock.setblocking(0)
            try:
                bombfuse.timeout(params.timeout, nonblock_handshake_proc)
            except bombfuse.TimeoutError as e:
                raise amncore.xcepts.ConnectionTimeout(params.peerhost, params.peerport)

    def write(self, buf, **kwargs):
        if self.allow_nonblock() == False:
            self.sslsock.setblocking(1)
            return self.sslsock.send(buf)

        total_sent = 0
        total_length = len(buf)
        block_size = 16384
        retry_time = 0.5
        
        self.sslsock.setblocking(0)
        while True:
            try:
                while (total_sent < total_length):
                    s = amncore.sync.select_s([], [self.sslsock], [], 0.25)
                    if s == ([],[],[]) or self.sslsock not in s[1]:
                        continue
                    data = buf[total_sent:block_size]
                    sent = self.sslsock.send(data)
                    if sent > 0:
                        total_sent += sent
                break
            except IOError as e:
                return None
            except OpenSSL.SSL.WantReadError as e:
                select.select([self.sslsock], [], [], retry_time)
                continue
            except OpenSSL.SSL.WantWriteError as e:
                select.select([], [self.sslsock], [], retry_time)
                continue
            except OpenSSL.SSL.Error as e:
                return None
            except Exception as e:
                raise e

        return total_sent

    def read(self, length = None, **kwargs):
        retval = None
        if self.allow_nonblock() == False:
            self.sslsock.setblocking(1)
            if length is None:
                length = 16384
            try:
                retval = self.sslsock.recv(length)
            except (IOError, OpenSSL.SSL.Error) as e:
                return None
            return str(retval.decode('latin-1'))

        self.sslsock.setblocking(0)
        while True:
            try:
                s = amncore.sync.select_s([self.sslsock], [], [], 0.10)
                if (s == ([],[],[]) or self.sslsock not in s[0]):
                    continue
                retval = self.sslsock.recv(length)
                break
            except IOError as e:
                return None
            except OpenSSL.SSL.WantReadError as e:
                select.select([self.sslsock], [], [], 0.5)
                continue
            except OpenSSL.SSL.WantWriteError as e:
                select.select([], [self.sslsock], [], 0.5)
                continue
            except OpenSSL.SSL.Error as e:
                return None
            except Exception as e:
                raise e
        return str(retval.decode('latin-1'))
                
    def close(self):
        try:
            self.sslsock.close()
        except Exception as e:
            pass
        return super(SslTcpSock, self).close()

    def shutdown(how = socket.SHUT_RDWR):
        pass

    @property
    def peer_cert(self):
        if self.sslsock is not None:
            return self.sslsock.get_peer_certificate()

    @property
    def peer_cert_chain(self):
        if self.sslsock is not None:
            return self.sslsock.get_peer_cert_chain()

    def recv(self, *args, **kwargs):
        raise RuntimeError("Invalid recv() call on SSL socket")

    def send(self, *args, **kwargs):
        raise RuntimeError("Invalid send() call on SSL socket")

    def allow_nonblock(self):
        import amncore.compat
        avail = (hasattr(self.sslsock, "setblocking") == True)
        if avail == True and amncore.compat.is_windows() == True:
            return True
        return False
        
class SslTcpServer(sockets.TcpServer, SecureSock):
    def __init__(self, *args, **kwargs):
        if hasattr(self, "sslctx") == False:
            self.sslctx = None
        return super(SslTcpServer, self).__init__(*args, **kwargs)

    @classmethod
    def create(self, **kwargs):
        kwargs['Proto'] = 'tcp'
        kwargs['Server'] = True
        kwargs['SSL'] = True
        return self.create_param(sockets.SocketParams.from_kwargs(**kwargs))

    @classmethod
    def create_param(self, param):
        param.proto = 'tcp'
        param.server = True
        param.ssl = True
        return sockets.create_param(param)

    def initsock(self, params = None):
        if params is not None and params.sslctx is not None and params.sslctx.__class__ == OpenSSL.SSL.Context:
            self.sslctx = params.sslctx
        else:
            self.sslctx = self.makessl(params)
        return super(SslTcpServer, self).initsock(params)

    def accept(self, **kwargs):
        try:
            r, _, _ = select.select([self.sock],[],[],0.5)
            if self.sock in r:
                _sock, addr = self.sock.accept()
            else:
                _sock = None
        except socket.error as e:
            _sock = None

        if _sock is None:
            return None

        try:
            ssl = OpenSSL.SSL.Connection(self.sslctx, _sock)
            ssl.set_accept_state()

            if self.allow_nonblock(ssl) == False:
                try:
                    ssl.setblocking(1)
                    ssl.do_handshake()
                except socket.timeout as e:
                    _sock.close()
                    raise OpenSSL.SSL.Error
            else:
                while True:
                    try:
                        ssl.setblocking(0)
                        ssl.do_handshake()
                        break
                    except OpenSSL.SSL.WantReadError as e:
                        select.select([ssl], [], [], 0.10)
                        continue
                    except OpenSSL.SSL.WantWriteError as e:
                        select.select([], [ssl], [], 0.10)
                        continue
                    except Exception as e:
                        raise e
    
            cls = SslTcpSock
            sock = cls(sock = _sock, sslsock = ssl)
            sock.sslctx = self.sslctx
            
            return sock
        except OpenSSL.SSL.Error as e:
            _sock.close()
            return None
            
def verify_cb(conn, cert, errno, depth, result):
    return True



import OpenSSL, re

class X509Cert(object):
    def __init__(self):
        pass
        
    @classmethod
    def parse_pem(self, ssl_cert):
        cert = None
        key = None
        chain = None
        
        certs = []
        
        mach = re.findall(r'-----BEGIN\s*[^\-]+-----+\r?\n[^\-]*-----END\s*[^\-]+-----\r?\n?', ssl_cert)
        if mach is not None:
            for pem in mach:
                mach2 = re.search(r'PRIVATE KEY', pem)
                mach3 = re.search(r'CERTIFICATE', pem)
                
                if mach2 is not None:
                    key = OpenSSL.crypto.load_privatekey(OpenSSL.crypto.FILETYPE_PEM, pem)
                elif mach3 is not None:
                    certs.append(OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, pem))
        
        if len(certs) > 0:
            cert = certs.pop(0)
            chain = certs
            
        return [key, cert, chain]
        
    @classmethod
    def parse_pem_file(self, ssl_cert_file):
        data = ''
        fd = open(ssl_cert_file, 'rb')
        data = fd.read()
        fd.close()
        
        return self.parse_pem(data)

    @classmethod
    def get_cert_hash(self, ssl_cert):
        hcert = self.parse_pem(ssl_cert)
        
        if hcert is None or hcert[0] is None or hcert[1] is None:
            raise TypeError("Could not parse a private key and certificate")
            
        return hcert[1].digest('sha1')
        
    @classmethod
    def get_cert_file_hash(self, ssl_cert_file):
        data = ''
        fd = open(ssl_cert_file, 'rb')
        data = fd.read()
        fd.close()
        
        return self.get_cert_hash(data)